// runAgent.js
// ------------------------------------------------------------------
//
// There is an API for this target.
//
// GET /status
//   returns a json payload providing status of the service.
//   Keep in mind the status is for the nodejs logic on a single MP only.
//   There are typically multiple MPs, so invoking GET /status multiple
//   times in succession will likely deliver different responses.
//
//   eg,
//     curl -i http://iloveapis2015-test.apigee.net/nagent/status
//
//
// POST /control
//   pass a x-www-form-urlencoded payload . eg, Use this header:
//        Content-type:application/x-www-form-urlencoded
//
//   The payload can be
//
//   action=start
//
//     -or-
//
//   action=stop
//
//   eg,
//     curl -i http://iloveapis2015-test.apigee.net/nagent/control -d 'action=stop'
//
//   You need to send this request just once, to stop all MPs
//   that are running, or to re-start all MPs.
//
//
// last saved: <2015-October-14 19:26:16>
// ------------------------------------------------------------------
//
// Copyright © 2015 Dino Chiesa and Apigee Corp
// All rights reserved.
//
// ------------------------------------------------------------------

var sleepTimeBetweenCyclesInMs = 2400,
    express = require('express'),
    app = express(),
    apigee = require('apigee-access'), // to put and get to and from the Edge cache
    cache = apigee.getCache(undefined, {scope: 'application'}), // get the default cache
    gStatusCacheKey = 'nagent-status', // should be unique among all cache keys
    gStatus = {
      version: 'Wednesday, 14 October 2015, 17:32',
      times : {
        start : (new Date()).toString(),
        lastRun : (new Date()).toString(),
      },
      nCycles : 0
    };


function copyHash(obj) {
  var copy = {};
  if (null !== obj && typeof obj == "object") {
    Object.keys(obj).forEach(function(attr){copy[attr] = obj[attr];});
  }
  return copy;
}

// =======================================================
//
// The simple API exposed by this script allows POST /control
// and GET /status
//
// =======================================================

app.use(express.urlencoded());

app.get('/status', function(request, response) {
  var payload;
  response.header('Content-Type', 'application/json');
  gStatus.times.current = (new Date()).toString();
  payload = copyHash(gStatus);
  cache.get(gStatusCacheKey, function(e, value) {
    if (e) {
      payload.error = true;
      payload.cacheException = e.toString();
      response.send(500, JSON.stringify(payload, null, 2) + "\n");
    }
    else {
      payload.cachedStatus = value || '-none-';
      response.send(200, JSON.stringify(payload, null, 2) + "\n");
    }
  });
});



/*
* controlling the app - for example start or stop - is a bit tricky
* because there are multiple nodejs processes running, yet only one of
* them receives the control request.  We therefore need a way of
* synchronizing state between the nodejs processes.  This code uses the
* Edge cache for that purpose.
*
* When a control request arrives, the handler writes to the cache.  All
* other instances of this process read from the cache in order to
* determine the state ("running", "stopped") they should adopt.
*
**/

app.post('/control', function(request, response) {
  var payload,
      // post body parameter, or query param
      action = request.body.action || request.query.action,
      loglevel = request.body.loglevel || request.query.loglevel,
      putCallback = function(e) {
        cache.get(gStatusCacheKey, function(e, value) {
          if (e) {
            payload.error = true;
            payload.cacheException = e.toString();
            response.send(500, JSON.stringify(payload, null, 2) + "\n");
          }
          else {
            payload.cachedStatus = value;
            response.send(200, JSON.stringify(payload, null, 2) + "\n");
          }
        });
      };

  response.header('Content-Type', 'application/json');
  payload = copyHash(gStatus);
  payload.times.current = (new Date()).toString();

  if (action != "stop" && action != "start") {
    payload.error = "unsupported request (action)";
    response.send(400, JSON.stringify(payload, null, 2) + "\n");
    return;
  }

  // there is a 2x2 matrix of possibilities.
  // state = {stopped, running}
  // action = {stop, start}
  // A different response is required for each combination.
  cache.get(gStatusCacheKey, function(e, value) {
    if (e) {
      payload.error = true;
      payload.cacheFail = true;
      payload.cacheException =  e.toString();
      response.send(500, JSON.stringify(payload, null, 2) + "\n");
    }
    else {
      payload.cachedStatus = value;
      if (value == "stopped") {
        if (action == "stop") {
          // nothing to do...send a 400.
          payload.error = "already stopped";
          response.send(400, JSON.stringify(payload, null, 2) + "\n");
        }
        else {
          // action == start
          console.log('starting...');
          cache.put(gStatusCacheKey, "running", 8640000, putCallback);
        }
      }
      else {
        // is marked "running" now.
        if (action == "stop") {
          console.log('stopping...');
          cache.put(gStatusCacheKey, "stopped", 8640000, putCallback);
        }
        else {
          // action == start
          // nothing to do, send a 400.
          payload.error = "already running";
          response.send(400, JSON.stringify(payload, null, 2) + "\n");
        }
      }
    }
  });
});


// for all other requests....
app.all(/^\/.*/, function(request, response) {
  response.header('Content-Type', 'application/json');
  response.send(404, '{ "message" : "This is not the server you\'re looking for." }\n');
});


// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

function runIteration() {
  try {
    cache.get(gStatusCacheKey, function(e, value) {
      if (e) {
        console.log('cannot retrieve status, presumed running.');
      }
      gStatus.cachedStatus = value || '-none-';
      if (e || value != "stopped") {
        // the agent is running
        gStatus.times.lastRun = (new Date()).toString();
        gStatus.nCycles++;
        console.log("Iteration " + gStatus.nCycles);
      }
      // else {
      //  // agent is stopped. do nothing during this cycle
      // }
    });
  }
  catch (exc1) {
    console.log("Exception:" + exc1);
    console.log(exc1.stack);
  }
  setTimeout(runIteration, sleepTimeBetweenCyclesInMs);
}

function kickoff() {
  cache.put(gStatusCacheKey, "running", 8640000, function(){});
  setTimeout(runIteration, 1); // 1ms
}

// xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


port = process.env.PORT || 5950;
app.listen(port, kickoff);
